/* Shopper
 * Copyright (C) 2008 David Greaves
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#ifndef SHOPPERLIST_H
#define SHOPPERLIST_H

#include <iostream>
#include <gtkmm.h>
#include <hildonmm.h>
#include <glib/gmarkup.h>
#include <list>
#include <set>
#include <map>
#include <signal.h>
using namespace std;

namespace Shopper
{
	// some forward declarations...
	class watchable;
	class List;
	class Category;
	class Item;


	// These are Shopper typedefs
	// FIXME: Is this the C++ way?
	typedef enum _List_action {
		CATEGORY_CHANGED,
		CATEGORY_DELETED,
		CATEGORY_LIST_CHANGED,
		ITEM_CHANGED,
		ITEM_DELETED,
		ITEM_LIST_CHANGED,
		LIST_CHANGED,
		LIST_STATE_CHANGED,
		LIST_CLEARED,
		LIST_ACTIVE_CAT_CHANGED
	} List_action;

	// a signal used between data and view
	typedef	sigc::signal<void, List_action, List*, Category*, Item*> signal_changed_t;

	// This is really application state but it needs saving to XML
	typedef enum sState_ {MAKING_LIST,
						  OUT_SHOPPING,
						  WHATS_LEFT} sState;  
		
////////////////////////////////////////////////////////////////
	class watchable
	{
	public:
		// actions emitted by signal
		signal_changed_t watch();       /* register for a changed signal */
		void pause();                   /* pause emitting change signals */
		void resume();                  /* resume emitting change signals */
	
		static Glib::ustring action_name(List_action a);  // for debugging
	
	protected:
		// Abstract class
		watchable();
		~watchable();
		virtual void emit_change(List_action a) =0 ; // must be implemented in each sub-class
		signal_changed_t signal_changed;
	};

////////////////////////////////////////////////////////////////
	class Category : public watchable
	{
	private:
		typedef enum sparseCategory_ {SPARSE} sparseCategory; // Only friends know about 
		Category(sparseCategory dummy);   // This constructor is for friends to instantiate objects
	public:
		Category();
		Category(const Category &src);
		Category(Glib::ustring name);
		~Category();

		void dbg();
		// accessors
		Glib::ustring print_id();
		int get_id();
		void set_id(int);
		void set_name(Glib::ustring);
		Glib::ustring get_name();
		friend bool operator< (const Category &a, const Category &b);
		friend bool cat_cmp (const Category *a, const Category *b);

	
		static const Category     EVERYTHING;   /* Used for filter */
	private:
		int            id;          /* The id of the item */
//		int            pos;         /* Allow user defined sorting */
		Glib::ustring  name;        /* The item name */
		static int     id_master;   /* Used for uniqueness */
		
		virtual void emit_change(Shopper::List_action);

		friend class ListParser;
		friend class XMLWriter;
	};

////////////////////////////////////////////////////////////////
	class Item : public watchable
	{
	private:
		typedef enum sparseItem_ {SPARSE} sparseItem; // Only friends know about 
		Item(sparseItem dummy);   // This constructor is for friends to instantiate objects

	public:
		Item();
		Item(Category &c, Glib::ustring desc, Glib::ustring note, gboolean wanted, gboolean bought);
		~Item();

		void dbg();
		
		// accessors
		Glib::ustring print_id();
		int get_id();
		void set_id(int);
		void set_desc(Glib::ustring);
		Glib::ustring get_desc();
		void set_note(Glib::ustring);
		Glib::ustring get_note();
		void set_wanted(gboolean);
		gboolean get_wanted();
		void set_bought(gboolean);
		gboolean get_bought();
		void set_category(Category *c);
		Category* get_category();
	
	private:
		int            id;           /* The id of the item (why?) */
		static int     id_master;    /* Used for uniqueness */
		Category      *category;     /* The id of the category the item is in */
		Glib::ustring  desc;         /* The item desc */
		Glib::ustring  note;         /* User note (flavour, quantity)*/
		gboolean       wanted;       /* do we want one of these? */
		gboolean       bought;       /* when out shopping, have we bought one of these */

		virtual void emit_change(Shopper::List_action);
		friend class ListParser;
		friend class XMLWriter;
	};
////////////////////////////////////////////////////////////////
	class List : public watchable
	{
		// declare some friends used to load/save Lists
		friend class ListParser;
		friend class XMLWriter;
	public:
		// Constructors
		List();
		List(Glib::ustring dump);   // initiate from a dump
		~List();
		static List* List::from_file(Glib::ustring filename);
		void List::to_file(Glib::ustring filename);

		void dbg();
		
		//List rawList();  // err dunno
		operator Glib::ustring();    // ustring cast gives XML
		// ostream (>>) gives XML
		friend std::ostream& operator<< (std::ostream &out, const List &l);
		
		void add(Item &it);         // List will manage Item
		void rm(Item &it);          // Delete item
		void add(Category &cat);    // List will manage Category
		void rm(Category &cat);     // Delete Category and all items
	private:
		// Helper functions used in managing category order
		void resequence();
		void reorder();
		void swap_categories(int, int);
	public:
		void swap_categories(Category *a, Category *b);

	public:
		// accessors
		Glib::ustring get_name();
		void set_name(Glib::ustring);
		sState get_state();
		void set_state(sState);

		// Filtering
		Category* get_active_category();
		gboolean is_category_active(Category &c);
		void no_category_active();
		void make_category_active(Category &c);
		void make_category_inactive(Category &c);
		void make_next_category_active();
		void make_prev_category_active();

		// Mode handling
		void boughtNothing();  // marks all items as not bought
		void newList();        // marks all items as not wanted  
		Glib::ustring List::modeText();
		static Glib::ustring List::modeText(sState s);

		// collection iterators
		typedef list<Item*>::const_iterator pItemIter;
		typedef list<Category*>::const_iterator pCategoryIter;
		pItemIter itemsI();
		pItemIter itemsEnd();
		pCategoryIter categoriesI();
		pCategoryIter categoriesEnd();
				
	private:
		void cycle_active_category(bool f);
		// members:
		Glib::ustring       name;
		sState              state;           // making, out:all, out:left		
		Category*           active_category ; 
// If using a set for multiple active categories
//		set<Category>      active_categories ; 
		map<int, Category*> cat_by_id;
		list<Category*>     categories;
		list<Item*>         items;

		virtual void emit_change(Shopper::List_action);

	};
////////////////////////////////////////////////////////////////
	class ListParser : public Glib::Markup::Parser
	{
		friend class List; // Given protected constructor, only Lists can make ListParsers
	protected:
		ListParser(List *l);   // Expected to be passed 'this'
//		~ListParser();
		void from_string(Glib::ustring);
		void start_list_el(Glib::Markup::ParseContext& context, const AttributeMap& attr);
		void start_cat_el(Glib::Markup::ParseContext& context, const AttributeMap& attr);
		void start_item_el(Glib::Markup::ParseContext& context, const AttributeMap& attr);
		void on_text (Glib::Markup::ParseContext& context, const Glib::ustring& el);
		void on_passthrough (Glib::Markup::ParseContext& context, const Glib::ustring& el);

		void 	on_start_element (Glib::Markup::ParseContext& context,
										  const Glib::ustring& element_name,
										  const AttributeMap& attributes);
		void 	on_end_element (Glib::Markup::ParseContext& context,
										const Glib::ustring& element_name);
		void 	on_error (Glib::Markup::ParseContext& context, const Glib::MarkupError& error);
		List*     l;
		Category* c;
		Item*     i;
		map<int, Category*> cat_by_id;
		int       active_cat_id;
	};
////////////////////////////////////////////////////////////////
	class XMLWriter
	{
	private:
		XMLWriter(const List *l);   // Expected to be passed 'this'
		std::ostream& write(std::ostream &o);
		friend class List; // Given a private constructor, only Lists can make XMLWriters
		friend std::ostream& operator<< (std::ostream &out, const List &l);
		const List *l;
	};
	
}

#endif //SHOPPERLIST_H



// Ordering
//
// This needs 2 functions: re-sequence and re-order
//
// re-sequence iterates the list in the current list order and sets
// 'id' incrementally
//
// re-order causes the list to examine all the member id's and
// re-order itself
//
// Each category has a sort position - 'id'.
// On load this is used to insert into the <list>
// post-load the list is re-sequenced
//
// On 'add' the category is inserted according to a probably duplicate id
// Then the list is re-sequenced.
//
// To re-order the list, the id's are swapped and the list is re-ordered.
// 
//

