/* Shopper
 * Copyright (C) 2008 David Greaves
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */
#include "shopper.h"           // automake, i8n, gettext
#include <glibmm.h>            // Glib C++ bindings
#include <gtkmm.h>             // Gtk C++ bindings
//#include <hildonmm.h>          // Hildon for C++
#include <hildon-fmmm.h>       // Hildon C++ filemanager
#include <gconfmm.h>           // Gnome::Conf
#include <iostream>            // IO Stream
#include <sstream>             // String Stream
#ifdef XRANDR
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/extensions/Xrandr.h>
#endif
#include "shopperList.h"       // Data objects
#include "shopperView.h"       // Main shopper items+list
#include "ItemDialog.h"        // Popup for editing Items
#include "CategoryDialog.h"    // Popup for editing Categories
#include "MainWindow.h"        // Me

//#define DEBUG(...) cerr << __VA_ARGS__
#define DEBUG(...)

using namespace Gtk;

MainWindow::MainWindow(Glib::ustring file) :
	lv(NULL),
	mylist(NULL),
	full_screen(false),
	categoryChooser(),
	filename(file),
	has_rotation(false)
{
	init_MainWindow(NULL);
}
MainWindow::MainWindow(Shopper::List &l) :
	lv(NULL),
	mylist(NULL),
	full_screen(false),
	categoryChooser(),
	filename(""),
	has_rotation(false)
{
	init_MainWindow(&l);
}
void MainWindow::init_MainWindow(Shopper::List *l)
{
	// The constructor handles only 'one off' stuff - main menu,
	// buttons etc. Anything List specific is seperated off to
	// handle_list() so it can be reused when opening new lists


	// xrandr detection from adv-backlight...
#ifdef XRANDR
	Display *dpy;
	int xrr_event_base, xrr_error_base;
	int xrr_major, xrr_minor;
	dpy = XOpenDisplay (NULL);
	if (dpy == NULL)
	{
		DEBUG("Couldn't open display\n");
		has_rotation = false;
	} else {
		if (XRRQueryExtension (dpy, &xrr_event_base, &xrr_error_base) == false
			|| XRRQueryVersion (dpy, &xrr_major, &xrr_minor) == 0 || xrr_major != 1 || xrr_minor < 1)
			has_rotation = false;
		else
			has_rotation = true;
		XCloseDisplay (dpy);
	}
#endif
	
	// Create actions for menus and toolbars:
	m_refActionGroup = ActionGroup::create();

	// Menus
	m_refActionGroup->add(Action::create("MainMenu", "ShopperMenu") );
	m_refActionGroup->add(Action::create("Item", _("Items")) ); //Sub-menu.
	m_refActionGroup->add(Action::create("ClearList", _("Clear List")) ); //Sub-menu.
	m_refActionGroup->add(Action::create("TickMode", _("Mode")) ); //Sub-menu.
	m_refActionGroup->add(Action::create("File", _("File")) ); //Sub-menu.

	// Add action
	m_refActionGroup->add(Action::create("AddItem", Stock::ADD, _("_Add Item"),
										 _("Add a new Item")),
						  sigc::mem_fun(*this, &MainWindow::on_action_add_item) );
	m_refActionGroup->add(Action::create("AddCategory", Stock::ADD, "_Add Category",
										 "Add a Category"),
						  sigc::mem_fun(*this, &MainWindow::on_action_add_category) );
	m_refActionGroup->add(Action::create("EditCategory", Stock::EDIT, "_Edit Category",
										 "Edit a Category"),
						  sigc::mem_fun(*this, &MainWindow::on_action_edit_category) );

	// Next/Prev category
	m_refActionGroup->add(Action::create("NextCategory", Stock::MEDIA_FORWARD),
						  sigc::mem_fun(*this, &MainWindow::on_action_next) );
//						  sigc::mem_fun(*mylist, &Shopper::List::make_next_category_active) );
	m_refActionGroup->add(Action::create("PrevCategory", Stock::MEDIA_REWIND),
						  sigc::mem_fun(*this, &MainWindow::on_action_prev) );
//						  sigc::mem_fun(*mylist, &Shopper::List::make_prev_category_active) );

	// Clear actions
	m_refActionGroup->add(Action::create("ClearWanted",	Stock::NEW, "_Clear Wanted",
										 "Clear Wanted"),
						  sigc::bind( sigc::mem_fun(*this, &MainWindow::on_action_clear_list), CLEAR_WANTED) );
	m_refActionGroup->add(Action::create("ClearBought",Stock::MISSING_IMAGE, "_Empty Basket",
										 "Empty Basket"),
						  sigc::bind( sigc::mem_fun(*this, &MainWindow::on_action_clear_list), CLEAR_BOUGHT) );

	// Shopping mode actions
	m_refActionGroup->add(Action::create("OutShoppingFull",	Stock::JUSTIFY_LEFT, "Out Shopping: _Full List",
										 "Show everything I need"),
						  sigc::bind( sigc::mem_fun(*this, &MainWindow::on_action_set_state), Shopper::OUT_SHOPPING ) );
	m_refActionGroup->add(Action::create("OutShoppingLeft", Stock::APPLY, "Out Shopping: What's _Left",
										 "What's left?"),
						  sigc::bind( sigc::mem_fun(*this, &MainWindow::on_action_set_state), Shopper::WHATS_LEFT) );
	m_refActionGroup->add(Action::create("MakingList", Stock::NEW, "_Making List",
										 "Select things to buy"),
						  sigc::bind( sigc::mem_fun(*this, &MainWindow::on_action_set_state), Shopper::MAKING_LIST ) );

	// File actions
	m_refActionGroup->add(Action::create("FileNew", Stock::NEW),
						  sigc::mem_fun(*this, &MainWindow::on_action_file_new) );
	m_refActionGroup->add(Action::create("FileOpen", Stock::OPEN),
						  sigc::mem_fun(*this, &MainWindow::on_action_file_open) );
	m_refActionGroup->add(Action::create("FileSave", Stock::SAVE),
						  sigc::mem_fun(*this, &MainWindow::on_action_file_save) );
	m_refActionGroup->add(Action::create("FileSaveAs", Stock::SAVE_AS),
						  sigc::mem_fun(*this, &MainWindow::on_action_file_saveas) );
	m_refActionGroup->add(Action::create("FileQuit", Stock::QUIT),
						  sigc::mem_fun(*this, &MainWindow::on_action_file_quit) );

	// Fullscreen
	m_refActionGroup->add(Action::create("FullScreen", Stock::FULLSCREEN),
						  sigc::mem_fun(*this, &MainWindow::on_action_fullscreen) );

	// Rotate
	m_refActionGroup->add(Action::create("Rotate", Stock::REFRESH),
						  sigc::mem_fun(*this, &MainWindow::on_action_rotate) );

	// Window close
	signal_delete_event().connect(sigc::mem_fun(*this, &MainWindow::on_signal_window_closed));

	// OK now build some UI for the actions
	m_refUIManager = UIManager::create();
	m_refUIManager->insert_action_group(m_refActionGroup);
  
	add_accel_group(m_refUIManager->get_accel_group());

	// Layout the actions in a menu and toolbar. In Hildon you have to
	// use "popup" for the menu because you can only have one menu and
	// cannot add a whole menubar.
  
	Glib::ustring ui_info = 
        "<ui>"
        "    <popup action='MainMenu'>"
        "      <menu action='Item'>"
	    "        <menuitem action='AddItem'/>"
	    "        <menuitem action='AddCategory'/>"
	    "        <menuitem action='EditCategory'/>"
        "      </menu>"
        "      <menu action='ClearList'>"
        "        <menuitem action='ClearWanted'/>"
        "        <menuitem action='ClearBought'/>"
        "      </menu>"
        "      <menu action='TickMode'>"
        "        <menuitem action='OutShoppingFull'/>"
        "        <menuitem action='OutShoppingLeft'/>"
        "        <menuitem action='MakingList'/>"
        "      </menu>"
        "      <separator/>"
        "      <menu action='File'>"
        "        <menuitem action='FileNew'/>"
        "        <menuitem action='FileOpen'/>"
        "        <menuitem action='FileSave'/>"
        "        <menuitem action='FileSaveAs'/>"
        "        <menuitem action='FileQuit'/>"
        "      </menu>"
        "    </popup>"
        "  <toolbar  name='ToolBar'>"
        "    <toolitem action='AddItem'/>"
        "    <toolitem action='FileSave'/>"
		"    <separator/>"
        "    <toolitem action='OutShoppingFull'/>"
        "    <toolitem action='OutShoppingLeft'/>"
        "    <toolitem action='MakingList'/>"
		"    <separator/>"
        "    <toolitem action='PrevCategory'/>"
		"    <separator expand='true' action='ButtonSpacer'/>"		
        "    <toolitem action='NextCategory'/>"
        "    <toolitem action='FullScreen'/>"
        "    <toolitem action='Rotate'/>"
        "  </toolbar>"
        "</ui>";

	std::auto_ptr<Glib::Error> ex;
	m_refUIManager->add_ui_from_string(ui_info, ex);
	if(ex.get())
	{ 
		DEBUG("building menus failed: " <<  ex->what());
	}
 
	// Get the menu and toolbar widgets, and add them to a container widget:
	Menu* pMenu = dynamic_cast<Menu*>(m_refUIManager->get_widget("/MainMenu"));
	if(pMenu)
		set_menu(*pMenu);

	Toolbar* pToolbar = dynamic_cast<Toolbar*>(m_refUIManager->get_widget("/ToolBar"));
	if(pToolbar)
		add_toolbar(*pToolbar);

	
	// add the dynamic menu to the toolbar before the 'ButtonSpacer'
	ToolItem* pTi = dynamic_cast<ToolItem*>(m_refUIManager->get_widget("/ToolBar/ButtonSpacer"));
	pToolbar->insert(categoryChooser, pToolbar->get_item_index(*pTi));

	// Remove the rotation button if there is no RandR
	if (! has_rotation) {
		ToolItem* pRot = dynamic_cast<ToolItem*>(m_refUIManager->get_widget("/ToolBar/Rotate"));
		pToolbar->remove(*pRot);
	}

	// If we got a filename then read it in...
	if (l == NULL) {
		l = Shopper::List::from_file(filename);
		if (l) { // If filename valid
			create_list_view(l);
		} else { // If filename wasn't read
			while (!mylist) { // mylist and list_view are done in a file_open()
				DEBUG("Couldn't open " << filename << " for reading\n");
				on_action_file_open();
				// FIXME :: Popup dialog box and allow new/open
			}
		}
	} else {
		// Now do List specific stuff
		create_list_view(l);
	}
	show_all();
}

MainWindow::~MainWindow()
{
}

void MainWindow::create_list_view(Shopper::List *list)
{
	// Remove the old listview... and any data
	remove();
	delete mylist;
	delete lv;

	// and use a list and a new main shopperList widget
	mylist = list;
	lv = new Shopper::ListView(*list);

	// GUI: set title, add the view, update the menu
	set_title(mylist->modeText());
	add_with_scrollbar(*lv);
	lv->show();
	// Make sure there's a menu to show to enable the popup functionality
	categoryChooser.Init(*mylist);
	show_all();
}

// Action handlers
// Add item
void MainWindow::on_action_add_item()
{
	DEBUG("An item may be added.\n");

	Shopper::ItemAdd itemD(mylist);
	if (itemD.run() == RESPONSE_ACCEPT) {
		DEBUG("\nADDED:\n");
	}	
}

void MainWindow::on_action_add_category()
{
	DEBUG("A cat may be added.\n");

	Shopper::CategoryAdd catD(mylist);
	if (catD.run() == RESPONSE_ACCEPT) {
		DEBUG("\nADDED:\n");
	}	
}
void MainWindow::on_action_edit_category()
{
	DEBUG("A cat may be edited.\n");

	Shopper::CategoryEdit catD(mylist, mylist->get_active_category());
	if (catD.run() == RESPONSE_ACCEPT) {
		DEBUG("\nADDED:\n");
	}	
}

// Category cycle
void MainWindow::on_action_next(){ 	mylist->make_next_category_active(); }
void MainWindow::on_action_prev(){ 	mylist->make_prev_category_active(); }

// Clear List
void MainWindow::on_action_clear_list(clear_list_mode m)
{
	for (Shopper::List::pItemIter it = mylist->itemsI(); it != mylist->itemsEnd(); it++) {
		(*it)->set_bought(false);  // Always clear bought
		if (m == CLEAR_WANTED) {
			(*it)->set_wanted(false);
		}
	}
	if (m == CLEAR_WANTED) {
		on_action_set_state(Shopper::MAKING_LIST);
	}
}
// Shopping mode
void MainWindow::on_action_set_state(Shopper::sState s)
{
	mylist->set_state(s);  // Set the state
	if (s == Shopper::MAKING_LIST) { // And go to category '1'
		DEBUG("Going to category 1 \n");
		mylist->make_category_active(**(mylist->categoriesI())); // FIXME - does the iterator use the right sort order?
	}
	if (s == Shopper::WHATS_LEFT || s == Shopper::OUT_SHOPPING) { // And go to category '1'
		DEBUG("Going to Everything\n");
		mylist->no_category_active();
	}
	DEBUG("Set title : " << mylist->modeText() <<"\n"); 
	set_title(mylist->modeText());
}
////////////////////////////////////////////////////////////////
// File Handling
void MainWindow::on_action_file_new()
{
	// FIXME: This really ought to check if the user *WANTS* to save...
	if (mylist)
		on_action_file_save();
	filename="";
	create_list_view(new Shopper::List(
		Glib::ustring(
			"<list name='Supermarket' state='0'>"
			"<category name='Aisle 1 - Dairy'>"
			"<item desc='Cheese' note='note, no smelly cheese' wanted='1' bought='0'/>"
			"</category>"
			"</list>"
			)));
}
void MainWindow::on_action_file_open()
{
	// Get a filename
	Hildon::FileChooserDialog dialog(*this, Gtk::FILE_CHOOSER_ACTION_OPEN);
	int response = dialog.run();
	dialog.hide();

	//Handle the response:
	if (response == Gtk::RESPONSE_OK) {
		// Open file and create a view from it.
		create_list_view(Shopper::List::from_file(filename=dialog.get_filename()));
		save_filename(filename); // FIXME this should be on success
	} else {
		if (mylist == NULL) on_action_file_new();  // If CANCEL pressed and there's no list, make a new one
	}
}
////////////////////////////////////////////////////////////////
void MainWindow::on_action_file_save()
{
	DEBUG("Saving....\n");
	if (filename=="") 
		on_action_file_saveas();
	else
		mylist->to_file(filename);
}
void MainWindow::on_action_file_saveas()
{
// Get a filename
	Hildon::FileChooserDialog dialog(*this, Gtk::FILE_CHOOSER_ACTION_SAVE);
	int response = dialog.run();
	dialog.hide();
	if (response == Gtk::RESPONSE_OK)
	{
		Glib::ustring filename = dialog.get_filename();
		save_filename(filename);
		mylist->to_file(filename);
    }
}
void MainWindow::on_action_file_quit()
{
	on_action_file_save();
	// FIXME
	hide(); //Closes the main window to stop the Gtk::Main::run().
}

bool MainWindow::on_signal_window_closed(GdkEventAny *ev)
{
	on_action_file_quit();
	return true; // propogate signal so hide() happens
}

//void MainWindow::on_action_hwkey()
//if (event->keyval == GDK_F6 && !keypressed){

void MainWindow::on_action_fullscreen()
{
	if (full_screen){
		unfullscreen();
	} else {
		fullscreen();
	}
	full_screen = !full_screen;
}
void MainWindow::save_filename(Glib::ustring name)
{
	static Glib::RefPtr<Gnome::Conf::Client>
		client = Gnome::Conf::Client::get_default_client();
	std::auto_ptr< Glib::Error > err;
	DEBUG("Saving " << name << " as list_filename\n");
	client->set("/apps/maemo/shopper/list_filename", name, err);
	if (err.get()){
		DEBUG("Couldn't set filename key " << err->what() << "\n");
	}
}
void MainWindow::on_action_rotate()
{
	if (! has_rotation) return;
	
	XRRScreenConfiguration *scr_config;
	Rotation current_rotation;
	gint size;
	int i;
	Display *dpy;
	gint screen;

	dpy = XOpenDisplay (NULL);
	screen = DefaultScreen (dpy);
	scr_config = XRRGetScreenInfo (dpy, RootWindow (dpy, screen));
	size = XRRConfigCurrentConfiguration (scr_config, &rotation);

	if (rotation == RR_Rotate_0) {
		rotation = RR_Rotate_90;
	} else {
		rotation = RR_Rotate_0; // always rotate back to 0
	}

	XRRSetScreenConfig (dpy, scr_config, RootWindow (dpy, screen), size, rotation, CurrentTime);
	XRRFreeScreenConfigInfo (scr_config);
	XCloseDisplay (dpy);

}
