/* Shopper
 * Copyright (C) 2008 David Greaves
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "shopper.h"           // automake, i8n, gettext
#include <glibmm.h>
#include <gtkmm.h>
#include <gtkmm/box.h>
#include <iostream>
#include "shopperList.h"
#include "ItemDialog.h"
#include "shopperView.h"

//#define DEBUG(...) cerr << __VA_ARGS__
#define DEBUG(...)

using namespace std;

namespace Shopper
{

// The widget shows the state of an item.
// Items are watched and when they emit a change signal, the display
// is updated (an old Tk/TCL idea)
	ItemView::ItemView(Item &it, List &l) :
		note_edit(NULL),
		myitem(&it),
		mylist(&l),
		desc(it.get_desc(), Gtk::ALIGN_LEFT),
		note(it.get_note(), Gtk::ALIGN_RIGHT)
	{
		// The tick represents either wanted or bought
		tick.set_active((l.get_state() == MAKING_LIST) ?
						it.get_wanted() : it.get_bought());
		// The labels are easy
//		desc.set_text(());
//		note.set_text(it.get_note());
		// Lay them out
		pack_start(tick, Gtk::PACK_SHRINK, 5);
		pack_start(outer_b, Gtk::PACK_EXPAND_WIDGET, 0);
		outer_b.pack_start(desc_b);
		desc_b.add(desc);
		outer_b.pack_start(note_b);
		note_b.add(note);

		// Create a menu (not so easy...)
		Glib::RefPtr<Gtk::ActionGroup> m_refActionGroup = Gtk::ActionGroup::create();
		m_refActionGroup->add( Gtk::Action::create("Pop", "EditMenu") );
		m_refActionGroup->add( Gtk::Action::create("MenuFile", "_File") );
		m_refActionGroup->add( Gtk::Action::create("Edit", Gtk::Stock::EDIT),
							   sigc::mem_fun(*this, &Shopper::ItemView::edit_item) );
		m_refActionGroup->add( Gtk::Action::create("Delete", Gtk::Stock::DELETE),
							   sigc::mem_fun(*this, &Shopper::ItemView::delete_item) );
		Glib::RefPtr<Gtk::UIManager> m_refUIManager =
			Gtk::UIManager::create();
		m_refUIManager->insert_action_group(m_refActionGroup);
		Glib::ustring ui_info =
			"<ui>"
			"  <popup action='Pop'>"
			"    <menuitem action='Edit'/>"
			"    <menuitem action='Delete'/>"
			"  </popup>"
			"</ui>";
		std::auto_ptr<Glib::Error> ex;
		m_refUIManager->add_ui_from_string(ui_info, ex);
		if(ex.get())
		{ 
			DEBUG("building menus failed: " <<  ex->what());
		}
 		Gtk::Widget* pMenu = m_refUIManager->get_widget("/Pop");

		// Respond to user clicking the checkbox.
		tick.signal_toggled()
			.connect( sigc::mem_fun(*this, &Shopper::ItemView::tick_clicked) );
		// And clicking the description is the same.
		desc_b.signal_button_press_event()
			.connect( sigc::mem_fun(*this, &Shopper::ItemView::desc_clicked) );

		// Respond to clicking the note
		note_b.signal_button_press_event()
			.connect( sigc::mem_fun(*this, &Shopper::ItemView::note_show_edit) );

		// Now connect the item's changed signal to this object's updater
		watch_item = it.watch()
			.connect( sigc::mem_fun(*this, &Shopper::ItemView::item_changed) );

		// And the watch the list so we can change visibility if needed
		watch_list = l.watch()
			.connect( sigc::mem_fun(*this, &Shopper::ItemView::list_changed) );

		// Finally connect the menu using the 'C' tap_and_hold interface
		GtkWidgetTapAndHoldFlags f;
		gtk_widget_tap_and_hold_setup(
			(GtkWidget*)(desc_b.gobj()),
			(GtkWidget*)(pMenu->gobj()), NULL, f);
	
		// Make the components visible
		tick.show();
		desc.show();
		note.show();
		desc_b.show();
		note_b.show();
		outer_b.show();
		// Make the container visible according to visibility logic
		DEBUG("ItemView created for" <<it.get_desc());
		show();
	}

	ItemView::~ItemView()
	{
		watch_item.disconnect();  // FIXME: not sure this is needed...
		watch_list.disconnect();
		delete note_edit;
		note_edit=NULL;
	}

// Only shows if matching item should be seen.
	void ItemView::show()
	{
		// FIXME : this shouldn't be needed.
		if (!mylist->is_category_active(*myitem->get_category())) {
			DEBUG(" category inactive, hide\n");
			Gtk::HBox::hide();
			return;
		}

		// Hide if not wanted
		if (((mylist->get_state() == WHATS_LEFT) || (mylist->get_state() == OUT_SHOPPING))
			&& ! myitem->get_wanted()) {
			DEBUG(" shopping, not wanted, hide\n");
			Gtk::HBox::hide();
			return;
		}		

		if ((mylist->get_state() == WHATS_LEFT) && myitem->get_bought()) {
			DEBUG(" what's left+bought, hide\n");
			Gtk::HBox::hide();
			return;
		}		
		DEBUG(" show\n");
		Gtk::HBox::show();
		return;
	}

	bool ItemView::note_show_edit(GdkEventButton* ev)
	{
		DEBUG("Pressed button " << ev->button << "\n");
		note_edit = new Gtk::Entry();
		note_edit->set_activates_default(false);
		note_edit->set_width_chars(15); // FIXME
		note_edit->set_text(myitem->get_note());
		// self-destruct on unfocus or activate
		note_edit->signal_activate()  // FIXME : Crashes. activate conflicts withe note_b.remove()
			.connect( sigc::mem_fun(*this,
									(void (Shopper::ItemView::*)())
									&Shopper::ItemView::note_show_label) );
		note_edit->signal_button_press_event()
			.connect( sigc::mem_fun(*this,
									(bool (Shopper::ItemView::*)(GdkEventButton*))
									&Shopper::ItemView::note_show_label) );
		note_edit->signal_focus_out_event()
			.connect( sigc::mem_fun(*this,
									(bool (Shopper::ItemView::*)(GdkEventFocus*))
									&Shopper::ItemView::note_show_label) );
		
		note_b.remove(); // remove label
		note_b.add(*note_edit);
		note_edit->show();
		note_edit->grab_focus();
	}
	bool ItemView::note_show_label(GdkEventFocus* ev)
	{
		if (! ev->in){
			note_show_label();
		}
		return true; // FIXME why true? Something to do with ev propogation?
	}
	bool ItemView::note_show_label(GdkEventButton* ev)
	{
		if (ev->button ==3 ){
			DEBUG("Middle click\n");
			note_show_label();
		}
		return true; // FIXME why true? Something to do with ev propogation?
	}
	void ItemView::note_show_label()
	{
		note_b.remove(); // remove note_edit
		note_b.add(note);
		note.show();
		myitem->set_note(note_edit->get_text()); // Setting the data
												 // will call an
												 // update on the note
		delete note_edit;
		note_edit=NULL;
	}
	void ItemView::edit_item()
	{
		ItemEdit itemD(mylist, myitem);
		itemD.run();
	}
	void ItemView::delete_item()
	{
		DEBUG("Delete Item\n");
		mylist->rm(*(this->myitem));
	}
	bool ItemView::desc_clicked(GdkEventButton* ev)
	{
		// Make the tick tock
		DEBUG("Desc pressed\n");
		tick.set_active(!tick.get_active());
	}
	void ItemView::tick_clicked()
	{
		gboolean t = tick.get_active();
		gboolean prior;
		// Suspend our data watch - we're making the change and don't need
		// to be notified
		DEBUG("Tick " << t <<"\n");
		prior = watch_item.block(true);
		switch (mylist->get_state()) {
		case OUT_SHOPPING:
			DEBUG("shopping ");
		case WHATS_LEFT:
			DEBUG("what's left\n");
			myitem->set_bought(t);
			break;
		case MAKING_LIST:
			DEBUG("Making list\n");
			myitem->set_wanted(t);
			break;
		default:
			DEBUG("State invalid\n");
		}
		show();
		watch_item.block(prior);
	}

// When the underlying data changes, 
	void ItemView::item_changed(List_action action, List *l, Category *c, Item *i)
	{
		DEBUG("Item changed");
		switch (action){
		case ITEM_CHANGED:
			DEBUG(" updating view\n");
			tick.set_active((mylist->get_state() == MAKING_LIST) ?
							myitem->get_wanted() : myitem->get_bought());
			desc.set_text(myitem->get_desc());
			note.set_text(myitem->get_note());
			if (! mylist->is_category_active(*(myitem->get_category()))) hide();
			break;
		case ITEM_DELETED:
			mylist=NULL;
		default:
			DEBUG("Unknown action : "
					   << Shopper::watchable::action_name(action) << "\n");
			break;
		}
	}
	void ItemView::list_changed(List_action action, List *l, Category *c, Item *i)
	{
		switch (action){
		case LIST_STATE_CHANGED:
			tick.set_active((mylist->get_state() == MAKING_LIST) ?
							myitem->get_wanted() : myitem->get_bought());
			DEBUG("checking vis ");
			show();
			break;
		default:
			break;
		}
	}


// This class displays and manages a list.
//
// It creates ItemViews dynamically
// When the mode changes it updates the display
// When the active category changes, it updates the display
// 
// class ListView: public Gtk::VBox
	
	ListView::ListView(List &l) :
		mylist(&l)
	{
		update();

		// Now connect the List's state changed signal to this object's updater
		watch_list = l.watch()
			.connect( sigc::mem_fun(*this, &Shopper::ListView::list_changed) );
		show();
	}

	ListView::~ListView()
	{
		std::map<int, ItemView*>::iterator  itemVsI;
		
		for(itemVsI = shownItemVs.begin();
			itemVsI != shownItemVs.end(); ++itemVsI)
		{
			delete itemVsI->second;
		}
	}

	void ListView::filter_by(Category &c)  // Only view certain categories
	{
		mylist->make_category_active(c);
	}

	void ListView::list_changed(List_action action, List *l, Category *c, Item *i)
	{
		switch (action){
		case LIST_STATE_CHANGED:
		case CATEGORY_LIST_CHANGED:
		case ITEM_LIST_CHANGED:
		case LIST_ACTIVE_CAT_CHANGED:
			update();
			break;
		default:
			DEBUG("list_changed: Unknown action : "
					   << Shopper::watchable::action_name(action) << "\n");
			break;
		}
	}

	void ListView::update()
	{
		// Scans all the items in the Shopper::List and creates or
		// removes them from the items map
		ItemView *w;
		DEBUG("Updating: active category is \n");
		for(Shopper::List::pCategoryIter catI = mylist->categoriesI();
			catI != mylist->categoriesEnd(); ++catI)
		{
			DEBUG("   " << (*catI)->get_name() << " " );
			if (mylist->is_category_active(**catI)) {
				DEBUG("  *");
				}
			DEBUG("\n");
		}
		// Delete all the existing itemviews and then clear the map
		// FIXME: do we even need a map now?
		for(std::map<const int, ItemView*>::iterator it = shownItemVs.begin() ; it!= shownItemVs.end(); it++)
		{
			delete it->second;
		}
		shownItemVs.clear();

		for(Shopper::List::pItemIter itemI = mylist->itemsI();
			itemI != mylist->itemsEnd(); ++itemI)
		{
			DEBUG("checking " << (*itemI)->get_desc());
//			std::map<int, ItemView*>::iterator found = shownItemVs.find((*itemI)->get_id());
			if (mylist->is_category_active(*(*itemI)->get_category()))
			{
				DEBUG(" ... active");
//				if (found != shownItemVs.end()) {
//					DEBUG(" ... found itemView\n");
//				} else {
					w = new ItemView(**itemI, *mylist);
					// Keep track of our child widgets
					shownItemVs.insert(pair<int,ItemView*>((*itemI)->get_id(), w));
					// Pack from the start...
					pack_start(*w, Gtk::PACK_SHRINK);
					DEBUG(" ... made new itemView\n");
//				}
			} else {
				DEBUG(" ... not selected");
//				if (found != shownItemVs.end()) {
//					DEBUG(" removing old itemView\n");
//					shownItemVs.erase((*itemI)->get_id());
//					delete found->second; // FIXME - does this unpack it?
//				} else {
				DEBUG("\n");
//				}
			}
		}
	}
}
